library nueli.core.config;

import 'dart:math';
import 'dart:ui';

import 'package:flutter/cupertino.dart';
import 'package:flutter/gestures.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';

import 'bucket.dart';
import 'data.dart';
import 'util.dart' show AdvanceList, AddList, SubtractList;

part 'config/decor.dart';
part 'config/inkwell.dart';
part 'config/layout.dart';
part 'config/photo.dart';
part 'config/radii.dart';
part 'config/tranx.dart';
part 'config/txt.dart';
part 'config/umbra.dart';
part 'config/x3d.dart';

/// ## [Config] ~ Widget Configuration Class.
/// Extend this class to create your own custom
/// configuration for widgets.
/// ```
abstract class Config<T extends Config<T, IN, OUT>, IN extends ConfigIN<IN, OUT>, OUT extends ConfigOUT<OUT, IN>>
    extends StatefulWidget {
  /// the reference of the instance of [IN]
  /// associated with this [Config].
  /// see: *[_fallback]*
  /// ```
  final Ref<IN> _iN = $();

  /// the fallback or default settings to be used
  /// by this [Config] during [build]
  /// ```
  final IN _fallback;

  /// subclass should call this constructor to pass
  /// in the basic instance of [IN] as *[_fallback]*
  /// ```
  Config({Key? key, required IN fallback})
      : _fallback = fallback,
        super(key: key);

  /// *@fixed* ~ set the [child] of the resulting
  /// widget
  /// ```
  T child(Widget? child) => resolve(child, #children);

  /// *@fixed* ~ set the [children] of the
  /// resulting widget
  /// ```
  T children(List<Widget>? children) => resolve(children, #children);

  /// *@fixed* ~ set the [axis] for [children]
  /// arrangement in the resulting widget
  /// ```
  T axis(Axis? axis) => resolve(axis, #axis);

  @mustCallSuper
  T resolve<Type>(data, Symbol tag) {
    return update((d) => d.resolve(data, tag));
  }

  @mustCallSuper
  T copy<Type>(IN iN) => update((d) => d.copy(iN));

  @mustCallSuper
  T update(IN Function(IN data) change) {
    return this as T.._iN.set(change(_iN.get ?? _fallback));
  }

  /// subclass should override this method to
  /// define the associated [ConfigOUT]. of the
  /// Call this method to get an instance of the
  /// associated [ConfigOUT] data.
  /// ```
  OUT call() => (_iN.get ?? _fallback)();

  T operator *(double operand) => update((data) => data * operand);

  T operator +(T operand) => update((data) => data + operand().iN);

  T operator -(T operand) => update((data) => data + operand().iN);

  @override
  ConfigState<T, IN, OUT> createState() => ConfigState();
}

class ConfigState<T extends Config<T, IN, OUT>, IN extends ConfigIN<IN, OUT>, OUT extends ConfigOUT<OUT, IN>>
    extends State<T> {
  @override
  Widget build(BuildContext context) => widget().build(context);
}

abstract class ConfigIN<T extends ConfigIN<T, OUT>, OUT extends ConfigOUT<OUT, T>>
    with Add<T>, Subtract<T>, Multiply<T, double> {
  final Symbol typeID;
  List<Widget>? children;
  Axis? axis;

  ConfigIN(this.typeID);

  @mustCallSuper
  T resolve(data, Symbol tag) {
    if (data == null) return this as T;
    if (data is Widget && tag == #children) children = [data];
    if (data is List<Widget> && tag == #children) children = data;
    if (data is Axis && tag == #axis) axis = data;
    return this as T;
  }

  @mustCallSuper
  T copy(T iN) {
    children = iN.children;
    axis = iN.axis;
    return this as T;
  }

  OUT call();
}

abstract class ConfigOUT<T extends ConfigOUT<T, IN>, IN extends ConfigIN<IN, T>> {
  final IN iN;
  const ConfigOUT(this.iN);

  List<Widget> get children => iN.children ?? [];

  Axis? get axis => iN.axis;

  @mustCallSuper
  Widget get child => wrap(children);

  @mustCallSuper
  Widget wrap(List<Widget> children) {
    if (children.isUnitary) return children.first;
    if (children.isNotEmpty) {
      return axis == Axis.vertical
          ? Column(children: children, mainAxisSize: MainAxisSize.min)
          : Row(children: children, mainAxisSize: MainAxisSize.min);
    }
    return const SizedBox();
  }

  Widget build(BuildContext context) => child;
}

abstract class ConfigTween<T extends Config<T, IN, OUT>, IN extends ConfigIN<IN, OUT>, OUT extends ConfigOUT<OUT, IN>>
    extends Tween<T> {
  ConfigTween({required T begin, required T end}) : super(begin: begin, end: end);

  T get defaultConfig;
}

class ConfigTransition<T extends Config<T, IN, OUT>, IN extends ConfigIN<IN, OUT>, OUT extends ConfigOUT<OUT, IN>>
    extends AnimatedWidget {
  final Widget? child;

  const ConfigTransition({
    Key? key,
    required Animation<T> config,
    this.child,
  }) : super(key: key, listenable: config);

  Animation<T> get config => listenable as Animation<T>;

  @override
  Widget build(BuildContext context) {
    return AnimatedBuilder(
      animation: listenable,
      builder: (c, child) => config.value.child(child),
      child: child,
    );
  }
}
