part of nueli.core.config;

Layout get $layout => Layout();

/// Layout ~ [LayoutWidget] configuration class.
class Layout extends Config<Layout, LayoutIN, LayoutOUT> {
  Layout({Key? key}) : super(key: key, fallback: LayoutIN());

  /// *@varies* ~ set size
  Layout size(double x, [double? y]) => width(x).height(y ?? x);

  /// *@varies* ~ set width
  Layout width(double width) => resolve(width, #width);

  /// *@varies* ~ set height
  Layout height(double height) => resolve(height, #height);

  /// *@varies* ~ set width scale factor
  Layout widthFactor(double factor) => resolve(factor, #widthFactor);

  /// *@varies* ~ set height scale factor
  Layout heightFactor(double factor) => resolve(factor, #heightFactor);

  /// *@varies* ~ set padding
  Layout padding(EdgeInsets? padding) => resolve(padding, #padding);

  /// *@varies* ~ set margin
  Layout margin(EdgeInsets? margin) => resolve(margin, #margin);

  /// *@varies* ~ set alignment
  Layout alignment(Alignment? alignment) => resolve(alignment, #alignment);

  /// *@varies* ~ set box constraints
  Layout constraints(BoxConstraints? constraints) => resolve(constraints, #constraints);

  /// *@shortcut* ~ expand size to maximum
  Layout get expand => size(double.infinity);

  /// *@shortcut* ~ expand width to maximum
  Layout get expandWidth => width(double.infinity);

  /// *@shortcut* ~ expand height to maximum
  Layout get expandHeight => height(double.infinity);

  /// *@shortcut* ~ align to top left corner
  Layout get alignStart => alignment(Alignment.topLeft);

  /// *@shortcut* ~ align to center
  Layout get alignCenter => alignment(Alignment.center);

  /// *@shortcut* ~ align to bottom right corner
  Layout get alignEnd => alignment(Alignment.bottomRight);
}

class LayoutIN extends ConfigIN<LayoutIN, LayoutOUT> {
  double? width, height, widthFactor, heightFactor;
  Alignment? alignment;
  EdgeInsets? padding, margin;

  BoxConstraints? constraints;

  LayoutIN() : super(#decor);

  @override
  LayoutIN operator *(double operand) {
    return LayoutIN().copy(this)
      ..width = this().width != null ? this().width! * operand : null
      ..height = this().height != null ? this().height! * operand : null
      ..widthFactor = this().widthFactor * operand
      ..heightFactor = this().heightFactor * operand
      ..alignment = this().alignment != null ? this().alignment! * operand : null
      ..padding = this().padding * operand
      ..margin = this().margin * operand;
  }

  @override
  LayoutIN operator -(LayoutIN operand) {
    return LayoutIN().copy(this)
      ..width = this().width != null ? this().width! - (operand().width ?? this().width!) : null
      ..height = this().height != null ? this().height! - (operand().height ?? this().height!) : null
      ..widthFactor = this().widthFactor - operand().widthFactor
      ..heightFactor = this().heightFactor - operand().heightFactor
      ..alignment = this().alignment != null ? this().alignment! - (operand().alignment ?? this().alignment!) : null
      ..padding = this().padding - operand().padding
      ..margin = this().margin - operand().margin;
  }

  @override
  LayoutIN operator +(LayoutIN operand) {
    return LayoutIN().copy(this)
      ..width = this().width != null ? this().width! + (operand().width ?? this().width!) : null
      ..height = this().height != null ? this().height! + (operand().height ?? this().height!) : null
      ..widthFactor = this().widthFactor + operand().widthFactor
      ..heightFactor = this().heightFactor + operand().heightFactor
      ..alignment = this().alignment != null ? this().alignment! + (operand().alignment ?? this().alignment!) : null
      ..padding = this().padding + operand().padding
      ..margin = this().margin + operand().margin;
  }

  @override
  LayoutIN resolve(data, tag) {
    if (data is double && tag == #width) width = data;
    if (data is double && tag == #height) height = data;
    if (data is double && tag == #widthFactor) widthFactor = data;
    if (data is double && tag == #heightFactor) heightFactor = data;
    if (data is Alignment && tag == #alignment) alignment = data;
    if (data is EdgeInsets && tag == #margin) margin = data;
    if (data is EdgeInsets && tag == #padding) padding = data;

    if (data is BoxConstraints && tag == #constraints) constraints = data;
    return super.resolve(data, tag);
  }

  @override
  LayoutIN copy(iN) {
    width = iN.width ?? width;
    height = iN.height ?? height;
    widthFactor = iN.widthFactor ?? widthFactor;
    heightFactor = iN.heightFactor ?? heightFactor;
    alignment = iN.alignment ?? alignment;
    margin = iN.margin ?? margin;
    padding = iN.padding ?? padding;

    constraints = iN.constraints ?? constraints;
    return super.copy(iN);
  }

  @override
  LayoutOUT call() => LayoutOUT(this);
}

class LayoutOUT extends ConfigOUT<LayoutOUT, LayoutIN> {
  double? width;
  double? height;

  LayoutOUT(LayoutIN iN) : super(iN) {
    width = iN.width;
    height = iN.height;
  }

  double get widthFactor => iN.widthFactor ?? 1;
  double get heightFactor => iN.heightFactor ?? 1;
  Alignment? get alignment => iN.alignment;
  EdgeInsets get padding => iN.padding ?? EdgeInsets.zero;
  EdgeInsets get margin => iN.margin ?? EdgeInsets.zero;
  BoxConstraints? get constraints => iN.constraints;

  double get shortestSide => min(width ?? 0, height ?? 0);

  double get longestSide => max(width ?? 0, height ?? 0);

  @override
  Widget wrap(children) {
    return Transform(
      transform: Matrix4.identity()..scale(widthFactor, heightFactor),
      alignment: iN.alignment,
      child: LayoutWidget(
        onChange: (size) {
          width = size.width;
          height = size.height;
        },
        child: Container(
          child: super.wrap(children),
          constraints: constraints,
          width: width,
          height: height,
          alignment: iN.alignment,
          padding: padding,
          margin: margin,
        ),
      ),
    );
  }
}

class LayoutTween extends ConfigTween<Layout, LayoutIN, LayoutOUT> {
  LayoutTween({required Layout begin, required Layout end}) : super(begin: begin, end: end);

  @override
  Layout get defaultConfig => Layout();
}

class LayoutTransition extends ConfigTransition<Layout, LayoutIN, LayoutOUT> {
  const LayoutTransition({Key? key, required Animation<Layout> decor, required Widget child})
      : super(key: key, config: decor, child: child);
}

class LayoutWidget extends SingleChildRenderObjectWidget {
  final Function(Size size) onChange;

  const LayoutWidget({
    Key? key,
    required this.onChange,
    required Widget child,
  }) : super(key: key, child: child);

  @override
  RenderObject createRenderObject(BuildContext context) {
    return LayoutRenderObject(onChange);
  }
}

class LayoutRenderObject extends RenderProxyBox {
  Size? oldSize;
  final Function(Size size) onChange;

  LayoutRenderObject(this.onChange);

  @override
  void performLayout() {
    super.performLayout();

    Size newSize = child!.size;
    if (oldSize == newSize) return;

    oldSize = newSize;
    WidgetsBinding.instance!.addPostFrameCallback((_) {
      onChange(newSize);
    });
  }
}
