part of nueli.core.config;

// Photo shortcuts
Photo get $photo => Photo();
Photo get $image => Photo();
Photo get $icon => Photo();

/// Photo ~ [Image] and [Icon] configuration class.
class Photo extends Config<Photo, PhotoIN, PhotoOUT> {
  Photo({Key? key}) : super(key: key, fallback: PhotoIN());

  /// *@varies* ~ set color filter
  Photo color(Color color) => resolve(color, #color).blendMode(BlendMode.srcIn);

  /// *@fixed* ~ set picture data
  Photo picture(Picture? picture) => resolve(picture, #picture);

  /// *@fixed* ~ set blend mode
  Photo blendMode(BlendMode? blendMode) => resolve(blendMode, #blendMode);

  /// *@fixed* ~ set box fit
  Photo boxFit(BoxFit? boxFit) => resolve(boxFit, #boxFit);

  /// *@config reference* ~ set inkwell
  Photo inkwell(Inkwell inkwell) => resolve(inkwell, #inkwell);

  /// *@config reference* ~ set layout
  Photo layout(Layout layout) => resolve(layout, #layout);

  /// *@from [Inkwell]* ~ set tap callback
  Photo onTap(Function()? onTap) => resolve(onTap, #onTap);

  /// *@from [Inkwell]* ~ set longPress callback
  Photo onLongPress(Function()? onLongPress) => resolve(onLongPress, #onLongPress);

  /// *@from [Layout]* ~ set size
  Photo size(double x, [double? y]) => width(x).height(y ?? x);

  /// *@from [Layout]* ~ set width
  Photo width(double width) => resolve(width, #width);

  /// *@from [Layout]* ~ set height
  Photo height(double height) => resolve(height, #height);

  /// *@from [Layout]* ~ set width scale factor
  Photo widthFactor(double factor) => resolve(factor, #widthFactor);

  /// *@from [Layout]* ~ set height scale factor
  Photo heightFactor(double factor) => resolve(factor, #heightFactor);

  /// *@from [Layout]* ~ set padding
  Photo padding(EdgeInsets? padding) => resolve(padding, #height);

  /// *@from [Layout]* ~ set layout
  Photo margin(EdgeInsets? margin) => resolve(margin, #margin);

  /// *@from [Layout]* ~ set alignment
  Photo alignment(Alignment? alignment) => resolve(alignment, #alignment);

  /// *@from [Layout]* ~ set box constraints
  Photo constraints(BoxConstraints? constraints) => resolve(constraints, #constraints);

  /// *@from [Layout]* ~ expand size to maximum
  Photo get expand => size(double.infinity);

  /// *@from [Layout]* ~ expand width to maximum
  Photo get expandWidth => width(double.infinity);

  /// *@from [Layout]* ~ expand height to maximum
  Photo get expandHeight => height(double.infinity);

  /// *@from [Layout]* ~ align to top left corner
  Photo get alignStart => alignment(Alignment.topLeft);

  /// *@from [Layout]* ~ align to center
  Photo get alignCenter => alignment(Alignment.center);

  /// *@from [Layout]* ~ align to bottom right corner
  Photo get alignEnd => alignment(Alignment.bottomRight);
}

class PhotoIN extends ConfigIN<PhotoIN, PhotoOUT> {
  Color? color;

  Picture? picture;
  BlendMode? blendMode;
  BoxFit? boxFit;

  Inkwell? inkwell;
  Layout? layout;

  PhotoIN() : super(#photo);

  @override
  PhotoIN operator *(double operand) {
    return PhotoIN().copy(this)
      ..color = this().color != null ? this().color! * operand : null
      ..inkwell = this().inkwell * operand
      ..layout = this().layout * operand;
  }

  @override
  PhotoIN operator -(PhotoIN operand) {
    return PhotoIN().copy(this)
      ..color = this().color != null ? this().color! - (operand().color ?? this().color!) : null
      ..inkwell = this().inkwell - operand().inkwell
      ..layout = this().layout - operand().layout;
  }

  @override
  PhotoIN operator +(PhotoIN operand) {
    return PhotoIN().copy(this)
      ..color = this().color != null ? this().color! + (operand().color ?? this().color!) : null
      ..inkwell = this().inkwell + operand().inkwell
      ..layout = this().layout + operand().layout;
  }

  @override
  PhotoIN resolve(data, tag) {
    if (data is Color && tag == #color) color = data;

    if (data is Picture && tag == #picture) picture = data;
    if (data is BlendMode && tag == #blendMode) blendMode = data;
    if (data is BoxFit && tag == #boxFit) boxFit = data;

    if (data is Inkwell && tag == #inkwell) inkwell = data;
    if (data is Layout && tag == #layout) layout = data;

    inkwell = this().inkwell.resolve(data, tag);
    layout = this().layout.resolve(data, tag);
    return super.resolve(data, tag);
  }

  @override
  PhotoIN copy(iN) {
    color = iN.color ?? color;

    picture = iN.picture ?? picture;
    blendMode = iN.blendMode ?? blendMode;
    boxFit = iN.boxFit ?? boxFit;

    inkwell = this().inkwell.copy(iN().inkwell().iN);
    layout = this().layout.copy(iN().layout().iN);
    return super.copy(iN);
  }

  @override
  PhotoOUT call() => PhotoOUT(this);
}

class PhotoOUT extends ConfigOUT<PhotoOUT, PhotoIN> {
  const PhotoOUT(PhotoIN iN) : super(iN);

  Color? get color => iN.color;

  Picture get picture => iN.picture ?? Picture.error;
  BlendMode get blendMode => iN.blendMode ?? BlendMode.srcIn;
  BoxFit get boxFit => iN.boxFit ?? BoxFit.contain;

  Inkwell get inkwell => iN.inkwell ?? $ink;
  Layout get layout => iN.layout ?? $layout;

  Picture get toPhotoData => iN.picture ?? Picture.library(CupertinoIcons.exclamationmark_circle);

  @override
  Widget build(context) {
    Widget result = child;

    /// library
    if (toPhotoData.data != null) {
      result = Icon(
        toPhotoData.data,
        color: color,
        size: layout().shortestSide,
      );
    }

    /// asset
    if (toPhotoData.path != null) {
      result = Image.asset(
        toPhotoData.path!,
        fit: boxFit,
        alignment: layout().alignment ?? Alignment.center,
        color: color,
        colorBlendMode: blendMode,
        width: layout().width,
        height: layout().height,
      );
    }

    /// network
    if (toPhotoData.src != null) {
      Widget loader(ImageChunkEvent chunk) {
        final loadValue =
            (chunk.expectedTotalBytes != null ? chunk.cumulativeBytesLoaded / chunk.expectedTotalBytes! : null);
        return Center(
          child: Container(
            height: 36,
            width: 36,
            padding: const EdgeInsets.all(8.0),
            child: CircularProgressIndicator(
              strokeWidth: 2,
              backgroundColor: palette.background,
              value: loadValue,
            ),
          ),
        );
      }

      Widget error([StackTrace? trace]) {
        return Center(
          child: Padding(
            padding: const EdgeInsets.all(8.0),
            child: $icon.picture(Picture.error),
          ),
        );
      }

      result = Image.network(
        toPhotoData.src!,
        color: color,
        fit: boxFit,
        alignment: layout().alignment ?? Alignment.center,
        colorBlendMode: blendMode,
        height: layout().height,
        width: layout().width,
        errorBuilder: (c, o, s) => error(s),
        loadingBuilder: (c, child, chunk) => chunk == null ? child : loader(chunk),
      );
    }
    return inkwell().wrap([layout().wrap([result,]), ...children]);
  }
}

class PhotoTween extends ConfigTween<Photo, PhotoIN, PhotoOUT> {
  PhotoTween({required Photo begin, required Photo end}) : super(begin: begin, end: end);

  @override
  Photo get defaultConfig => Photo();
}

class PhotoTransition extends ConfigTransition<Photo, PhotoIN, PhotoOUT> {
  const PhotoTransition({Key? key, required Animation<Photo> photo, required Widget child})
      : super(key: key, config: photo, child: child);
}
