part of nueli.core.config;

/// Tranx shortcut
Tranx get $tranx => Tranx();

/// Tranx ~ [AnimatedWidget] and [Tween] configuration class
/// All commands are fixed and instantaneous.
/// ```
class Tranx extends Config<Tranx, TranxIN, TranxOUT> {
  Tranx({Key? key}) : super(key: key, fallback: TranxIN());

  Tranx animate({
    Tween<double>? scale,
    Tween<Offset>? translate,
    Tween<double>? height,
    Tween<double>? width,
    Tween<double>? rotation,
    Tween<double>? opacity,
    Tween<Config>? configTween,
  }) {
    if (scale != null) resolve(_Tween.tween(_TweenType.scale, scale), #tween);
    if (translate != null) resolve(_Tween.tween(_TweenType.translate, translate), #tween);
    if (height != null) resolve(_Tween.tween(_TweenType.height, height), #tween);
    if (width != null) resolve(_Tween.tween(_TweenType.width, width), #tween);
    if (rotation != null) resolve(_Tween.tween(_TweenType.rotation, rotation), #tween);
    if (opacity != null) resolve(_Tween.tween(_TweenType.opacity, opacity), #tween);
    if (configTween != null) resolve(configTween, #tween);
    return this;
  }

  Tranx height(double begin, double end, [Alignment alignment = Alignment.center]) {
    return animate(width: Tween(begin: begin, end: end));
  }

  Tranx width(double begin, double end, [Alignment alignment = Alignment.center]) {
    return animate(height: Tween(begin: begin, end: end));
  }

  Tranx size(Size begin, Size end, [Alignment alignment = Alignment.center]) {
    return height(begin.width, end.width).width(begin.height, end.height);
  }

  Tranx scale(double begin, double end, [Alignment alignment = Alignment.center]) {
    return animate(scale: Tween(begin: begin, end: end));
  }

  Tranx translate(Offset begin, Offset end) => animate(translate: Tween(begin: begin, end: end));

  Tranx rotate(double begin, double end) => animate(rotation: Tween(begin: begin, end: end));

  Tranx fade(double begin, double end) => animate(opacity: Tween(begin: begin, end: end));

  Tranx decor(Decor begin, Decor end) => animate(configTween: DecorTween(begin: begin, end: end));

  Tranx ink(Inkwell begin, Inkwell end) => animate(configTween: InkwellTween(begin: begin, end: end));

  Tranx media(Photo begin, Photo end) => animate(configTween: PhotoTween(begin: begin, end: end));

  Tranx radii(Radii begin, Radii end) => animate(configTween: RadiiTween(begin: begin, end: end));

  Tranx umbra(Umbra begin, Umbra end) => animate(configTween: UmbraTween(begin: begin, end: end));

  Tranx txt(Txt begin, Txt end) => animate(configTween: TxtTween(begin: begin, end: end));

  Tranx x3D(X3D begin, X3D end) => animate(configTween: X3DTween(begin: begin, end: end));

  Tranx animation(Animation<double>? animation) => resolve(animation, #animation);
}

class TranxIN extends ConfigIN<TranxIN, TranxOUT> {
  List<Tween>? tween;
  Animation<double>? animation;
  Curve? curve;
  Curve? reverseCurve;
  Duration? duration;
  Duration? reverseDuration;

  TranxIN() : super(#tranx);

  @override
  TranxIN resolve(data, tag) {
    if (data is Tween && tag == #tween) tween = this().tween..add(data);
    if (data is Animation<double> && tag == #animation) animation = data;
    if (data is Curve && tag == #curve) curve = data;
    if (data is Curve && tag == #reverseCurve) reverseCurve = data;
    if (data is Duration && tag == #duration) duration = data;
    if (data is Duration && tag == #reverseDuration) reverseDuration = data;
    return super.resolve(data, tag);
  }

  @override
  @Deprecated('')
  TranxIN operator *(double operand) => this;

  @override
  @Deprecated('')
  TranxIN operator +(TranxIN operand) => copy(operand);

  @override
  @Deprecated('')
  TranxIN operator -(TranxIN operand) => copy(operand);

  @override
  TranxIN copy(TranxIN iN) {
    tween = iN.tween ?? tween;
    animation = iN.animation ?? animation;
    curve = iN.curve ?? curve;
    reverseCurve = iN.reverseCurve ?? reverseCurve;
    duration = iN.duration ?? duration;
    reverseDuration = iN.reverseDuration ?? reverseDuration;
    return super.copy(iN);
  }

  @override
  TranxOUT call() => TranxOUT(this);
}

class TranxOUT extends ConfigOUT<TranxOUT, TranxIN> implements Animation<double> {
  TranxOUT(TranxIN iN) : super(iN);

  List<Tween> get tween => iN.tween ?? [];

  Animation<double>? get animation => iN.animation;

  Curve get curve => iN.curve ?? Curves.easeOutExpo;

  Curve get reverseCurve => iN.reverseCurve ?? curve.flipped;

  Duration get duration => iN.duration ?? const Duration(milliseconds: 500);

  Duration get reverseDuration => iN.reverseDuration ?? duration;

  @override
  Widget build(context) {
    Widget child = this.child;
    if (animation == null) return child;
    Animation<double> anim = CurvedAnimation(parent: animation!, curve: curve);
    for (Tween tween in tween) {
      if (tween is _Tween<double>) {
        if (tween.type == _TweenType.scale) {
          child = ScaleTransition(scale: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.height) {
          child = SizeTransition(axis: Axis.vertical, sizeFactor: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.width) {
          child = SizeTransition(axis: Axis.horizontal, sizeFactor: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.opacity) {
          child = FadeTransition(opacity: tween.animate(anim), child: child);
        }
        if (tween.type == _TweenType.rotation) {
          child = RotationTransition(turns: tween.animate(anim), child: child);
        }
      }
      if (tween is _Tween<Offset>) {
        if (tween.type == _TweenType.translate) {
          child = SlideTransition(position: tween.animate(anim), child: child);
        }
      }
      if (tween is DecorTween) child = DecorTransition(decor: tween.animate(anim), child: child);
      if (tween is InkwellTween) child = InkwellTransition(inkwell: tween.animate(anim), child: child);
      if (tween is PhotoTween) child = PhotoTransition(photo: tween.animate(anim), child: child);
      if (tween is RadiiTween) child = RadiiTransition(radii: tween.animate(anim), child: child);
      if (tween is UmbraTween) child = UmbraTransition(umbra: tween.animate(anim), child: child);
      if (tween is TxtTween) child = TxtTransition(txt: tween.animate(anim), child: child);
      if (tween is X3DTween) child = X3DTransition(x3D: tween.animate(anim), child: child);
    }
    return child;
  }

  @override
  void addListener(VoidCallback listener) {
    animation!.addListener(listener);
  }

  @override
  void addStatusListener(AnimationStatusListener listener) {
    animation!.addStatusListener(listener);
  }

  @override
  Animation<U> drive<U>(Animatable<U> child) => animation!.drive(child);

  @override
  bool get isCompleted => animation!.isCompleted;

  @override
  bool get isDismissed => animation!.isDismissed;

  @override
  void removeListener(VoidCallback listener) {
    animation!.removeListener(listener);
  }

  @override
  void removeStatusListener(AnimationStatusListener listener) {
    animation!.removeStatusListener(listener);
  }

  @override
  AnimationStatus get status => animation!.status;

  @override
  String toStringDetails() => animation!.toStringDetails();

  @override
  double get value => animation!.value;
}

enum _TweenType { scale, translate, height, width, rotation, opacity }

class _Tween<T> extends Tween<T> {
  final _TweenType type;
  _Tween(this.type, {T? begin, T? end}) : super(begin: begin, end: end);
  _Tween.tween(this.type, Tween<T> tween) : super(begin: tween.begin, end: tween.end);
}
