part of nueli.core.config;

// Ace shortcuts
Umbra get $umbra => Umbra();
Umbra get $shadow => Umbra();

/// ## [Umbra] ~ [BoxShadow] configuration class.
/// Gives you access to configure a [BoxShadow].
/// ```
class Umbra extends Config<Umbra, UmbraIN, UmbraOUT> {
  Umbra({Key? key}) : super(key: key, fallback: UmbraIN());

  /// *@varies* ~ set the color of the resulting
  /// shadow.
  /// ```
  Umbra color(Color color, [int? alpha]) {
    return resolve(color.withAlpha(alpha ?? palette.shade.alpha), #umbraColor);
  }

  /// *@varies* ~ set the offset ([x], [y]) of the
  /// resulting shadow.
  /// ```
  Umbra offset(double x, [double? y]) {
    return resolve(Offset(x, y ?? x), #offset);
  }

  /// *@varies* ~ set the spread radius of the
  /// resulting shadow.
  /// ```
  Umbra spread(double spread) => resolve(spread, #spread);

  /// *@varies* ~ set the blur radius of the
  /// resulting shadow.
  /// ```
  Umbra blur(double blur) => resolve(blur, #blur);

  /// *@fixed* ~ set the decoration position of
  /// the resulting shadow.
  /// ```
  Umbra renderPosition(DecorationPosition? position) {
    return resolve(position, #position);
  }

  /// *@template* ~ create a new umbra configured to
  /// have a raised effect.
  /// ```
  factory Umbra.raised(double value) {
    return Umbra()
      ..blur(value / 6)
      ..spread(-0.5)
      ..offset(value, value);
  }

  /// *@template* ~ create a new umbra configured to
  /// have an elevated effect.
  /// ```
  factory Umbra.elevated(double value) {
    return Umbra()
      ..blur(value)
      ..spread(-0.27)
      ..offset(value / 8, value / 8);
  }
}

class UmbraIN extends ConfigIN<UmbraIN, UmbraOUT> {
  /// *@varies* ~ color of the shadow.
  /// ```
  Color? color;

  /// *@varies* ~ offset of the shadow.
  /// ```
  Offset? offset;

  /// *@varies* ~ spread radius of the shadow.
  /// ```
  double? spread;

  /// *@varies* ~ blur radius of the shadow.
  /// ```
  double? blur;

  /// *@varies* ~ position of the shadow.
  /// ```
  DecorationPosition? position;

  UmbraIN() : super(#umbra);

  @override
  UmbraIN operator *(double operand) {
    return UmbraIN().copy(this)
      ..color = this().color * operand
      ..offset = this().offset * operand
      ..spread = this().spread * operand
      ..blur = this().blur * operand;
  }

  @override
  UmbraIN operator -(UmbraIN operand) {
    return UmbraIN().copy(this)
      ..color = this().color - operand().color
      ..offset = this().offset - operand().offset
      ..spread = this().spread - operand().spread
      ..blur = this().blur - operand().blur;
  }

  @override
  UmbraIN operator +(UmbraIN operand) {
    return UmbraIN().copy(this)
      ..color = this().color + operand().color
      ..offset = this().offset + operand().offset
      ..spread = this().spread + operand().spread
      ..blur = this().blur + operand().blur;
  }

  @override
  UmbraIN resolve(data, tag) {
    if (data is Color && tag == #shadowColor) color = data;
    if (data is double && tag == #blur) blur = data;
    if (data is double && tag == #spread) spread = data;
    if (data is Offset && tag == #offset) offset = data;

    if (data is DecorationPosition && tag == #position) position = data;
    return super.resolve(data, tag);
  }

  @override
  UmbraIN copy(UmbraIN iN) {
    color = iN.color ?? color;
    blur = iN.blur ?? blur;
    spread = iN.spread ?? spread;
    offset = iN.offset ?? offset;

    position = iN.position ?? position;
    return super.copy(iN);
  }

  @override
  UmbraOUT call() => UmbraOUT(this);
}

class UmbraOUT extends ConfigOUT<UmbraOUT, UmbraIN> {
  const UmbraOUT(UmbraIN iN) : super(iN);

  Color get color => iN.color ?? palette.none;
  Offset get offset => iN.offset ?? Offset.zero;
  double get spread => iN.spread ?? 0;
  double get blur => iN.blur ?? 0;

  DecorationPosition get position {
    return iN.position ?? DecorationPosition.background;
  }

  BoxShadow get toBoxShadow {
    return BoxShadow(
      color: color,
      offset: offset,
      blurRadius: blur,
      spreadRadius: spread,
    );
  }

  List<BoxShadow> get toBoxShadowDeep => [toBoxShadow];

  @override
  Widget build(BuildContext context) {
    return DecoratedBox(
      decoration: BoxDecoration(boxShadow: toBoxShadowDeep),
      child: child,
      position: position,
    );
  }
}

class UmbraTween extends ConfigTween<Umbra, UmbraIN, UmbraOUT> {
  UmbraTween({required Umbra begin, required Umbra end}) : super(begin: begin, end: end);

  @override
  Umbra get defaultConfig => Umbra();
}

class UmbraTransition extends ConfigTransition<Umbra, UmbraIN, UmbraOUT> {
  const UmbraTransition({Key? key, required Animation<Umbra> umbra, required Widget child})
      : super(key: key, config: umbra, child: child);
}
