part of nueli.core.config;

// X3D shortcuts
X3D get $3D => X3D();

/// X3D ~ [Matrix4] configuration class.
class X3D extends Config<X3D, X3DIN, X3DOUT> {
  X3D({Key? key}) : super(key: key, fallback: X3DIN());

  X3D transform(TransformData transform) {
    return resolve(this().transforms..add(transform), #transforms);
  }

  X3D alignment(Alignment alignment) {
    return resolve(alignment, #alignment);
  }

  X3D skew({double x = 0, double y = 0, double z = 0}) => (skewX(x)..skewY(y))..rotate(z);
  X3D zoom(double zoom) => transform(TransformData(zoom, (t, d) => t..rotateZ(d)));
  X3D translate({double x = 0, double y = 0, double z = 0}) => (panX(x)..panY(y))..panZ(z);
  X3D shift({double x = 0, double y = 0, double z = 0}) => (offsetX(x)..offsetY(y))..offsetZ(z);
  X3D scale({double x = 0, double y = 0, double z = 0}) => (scaleX(x)..scaleY(y))..scaleZ(z);

  X3D skewX(double x) => transform(TransformData(x, (t, d) => t..rotateX(d)));
  X3D skewY(double y) => transform(TransformData(y, (t, d) => t..rotateY(d)));
  X3D rotate(double rotation) => transform(TransformData(rotation, (t, d) => t..rotateZ(d)));
  X3D panX(double x) => transform(TransformData(x, (t, d) => t..translate(d)));
  X3D panY(double y) => transform(TransformData(y, (t, d) => t..translate(0, d)));
  X3D panZ(double z) => transform(TransformData(z, (t, d) => t..translate(0, 0, d)));
  X3D offsetX(double x) => transform(TransformData(x, (t, d) => t..leftTranslate(d)));
  X3D offsetY(double y) => transform(TransformData(y, (t, d) => t..leftTranslate(0, d)));
  X3D offsetZ(double z) => transform(TransformData(z, (t, d) => t..leftTranslate(0, 0, d)));
  X3D scaleX(double x) => transform(TransformData(x, (t, d) => t..scale(d)));
  X3D scaleY(double y) => transform(TransformData(y, (t, d) => t..scale(0, d)));
  X3D scaleZ(double z) => transform(TransformData(z, (t, d) => t..scale(0, 0, d)));

  X3D base(Matrix4? base) {
    return resolve(base, #base);
  }

  factory X3D.perspective(int level) {
    return X3D().base(Matrix4.identity()..setEntry(3, 2, 0.001 * level));
  }
}

class X3DIN extends ConfigIN<X3DIN, X3DOUT> {
  Alignment? alignment;
  List<TransformData>? transforms;
  Matrix4? base;

  X3DIN() : super(#x3D);

  @override
  X3DIN operator *(double operand) {
    alignment = this().alignment * operand;
    transforms = this().transforms.map((e) => e * operand).toList();
    return this;
  }

  @override
  X3DIN operator -(X3DIN operand) {
    transforms = this().transforms.minus(operand().transforms);
    alignment = this().alignment - this().alignment;
    return this;
  }

  @override
  X3DIN operator +(X3DIN operand) {
    transforms = this().transforms.plus(operand().transforms);
    alignment = this().alignment + operand().alignment;
    return this;
  }

  @override
  X3DIN resolve(data, tag) {
    if (data is Alignment && tag == #alignment) alignment = data;
    if (data is List<TransformData> && tag == #transforms) transforms = data;

    if (data is Matrix4 && tag == #base) base = data;
    return super.resolve(data, tag);
  }

  @override
  X3DIN copy(X3DIN iN) {
    alignment = iN.alignment ?? alignment;
    transforms = iN.transforms ?? transforms;
    base = iN.base ?? base;
    return super.copy(iN);
  }

  @override
  X3DOUT call() => X3DOUT(this);
}

class X3DOUT extends ConfigOUT<X3DOUT, X3DIN> {
  X3DOUT(X3DIN iN) : super(iN);

  Alignment get alignment => iN.alignment ?? Alignment.center;
  List<TransformData> get transforms => iN.transforms ?? [];
  Matrix4 get base => iN.base ?? Matrix4.identity()
    ..setEntry(3, 2, 0.002);

  Matrix4 get toMatrix4 {
    Matrix4 mat = base;
    for (TransformData transform in transforms) {
      mat = transform(mat);
    }
    return mat;
  }

  @override
  Widget build(context) {
    return Transform(
      child: child,
      alignment: alignment,
      transform: toMatrix4,
    );
  }
}

class X3DTween extends ConfigTween<X3D, X3DIN, X3DOUT> {
  X3DTween({required X3D begin, required X3D end}) : super(begin: begin, end: end);

  @override
  X3D get defaultConfig => X3D();
}

class X3DTransition extends ConfigTransition<X3D, X3DIN, X3DOUT> {
  const X3DTransition({Key? key, required Animation<X3D> x3D, required Widget child})
      : super(key: key, config: x3D, child: child);
}

class TransformData with Subtract<TransformData>, Add<TransformData> {
  double value;
  Matrix4 Function(Matrix4, double) transform;
  TransformData(this.value, this.transform);
  Matrix4 call(Matrix4 matrix) => transform(matrix, value);

  @override
  TransformData operator -(TransformData operand) {
    return TransformData(value - operand.value, (mat, d) => transform(operand.transform(mat, d), d));
  }

  @override
  TransformData operator +(TransformData operand) {
    return TransformData(value + operand.value, (mat, d) => transform(operand.transform(mat, d), d));
  }

  TransformData operator *(double factor) {
    return TransformData(value + factor, (mat, d) => transform(mat, d));
  }
}
